

#include <algorithm>
#include <memory> // std::unique_ptr<>
#include <string.h> // For strcpy_s
#include <vector>
#include <map>
#include "inc/opennurbs.h"
#include "atf_tspline_user_data.h"

using namespace std;

// {10D2CFF0-8191-4cc7-AFDF-D8FAF16E2120}
GUID SimpleTMeshUserData::m_uuid = { 0x10d2cff0, 0x8191, 0x4cc7, { 0xaf, 0xdf, 0xd8, 0xfa, 0xf1, 0x6e, 0x21, 0x20 } };

ON_OBJECT_IMPLEMENT(SimpleTMeshUserData, ON_UserData, "10D2CFF0-8191-4cc7-AFDF-D8FAF16E2120");

SimpleTMeshUserData::SimpleTMeshUserData()
{
    m_userdata_uuid = m_uuid;
    // If m_userdata_copycount is 0, user data is not copied when object is copied.
    // if > 0, user data is copied and m_copycount is incremented when parent object
    // is copied.  The user data's operator=() is used to copy.
    m_userdata_copycount = 1;

    static const GUID TSplinesPlugIn_UUID =
    { 0xFD3C34A2, 0x64E4, 0x4124, { 0x9C, 0x48, 0xD3, 0xDE, 0x98, 0x65, 0x7A, 0x35 } };

    ON_UserData::m_application_uuid = TSplinesPlugIn_UUID;

    m_saved_in_smooth_mode = false;
}

bool SimpleTMeshUserData::GetDescription( ON_wString &s )
{
    s = L"T-Mesh Userdata";
    return TRUE;
}

const SimpleTMeshUserData *SimpleTMeshUserData::Get( const ON_Geometry *mesh_or_brep )
{
    if ( !mesh_or_brep )
        return nullptr;

    return SimpleTMeshUserData::Cast( mesh_or_brep->GetUserData(m_uuid) );
}

bool SimpleTMeshUserData::Archive() const
{
    return TRUE;
}

bool SimpleTMeshUserData::Read( ON_BinaryArchive& file )
{
    try
    {
        // Read the TSM block, which is a String followed by a Bool.
        // This block is present in all T-splines 3dm files.  However,
        // if the Bool value cannot be read, it is because there exists
        // no additional user data.  In that case, the Version read below
        // should also fail.
        bool end_reached = false;
        if( !ReadLegacyBlock(file, end_reached) )
        {
            // "Invalid format: could not read legacy TSM data block."
            m_rstatus_stack.Append( STMRR_ERROR_COULDNT_READ_LEGACY_BLOCK );
            return FALSE;
        }

        // HACK: Read a bool to ensure we're still ok
        // reading data.  For some reason, ReadInt()
        // succeeds even when there's no more data to read,
        // while ReadBool() seems to behave correctly.
        if( !end_reached )
        {
            if( !file.ReadBool(&end_reached) )
                end_reached = true;
        }

        if( !end_reached )
        {
            int version = 0;
            if( file.ReadInt(&version) )
            {
                switch( version )
                {
                case 1:
                    {
                        // T-Splines Plug-In v2.0+
                        bool read01_success = ReadVersion01( file );
                        if( !read01_success )
                            m_rstatus_stack.Append( STMRR_ERROR_READING_BLOCK_STRUCTURE );
                        return read01_success;
                    }
                    break;
                default:
                    // "Invalid format: unknown version number"
                    m_rstatus_stack.Append( STMRR_ERROR_UNKNOWN_BLOCK_STRUCTURE );
                    return FALSE;
                    break;
                }
            }
        }
    }

    //This catch-all is Okay since the function is not supposed throw exceptions in order to 
    //keep other parts of the process to continue. It relies on the return code as the error 
    //mechanism and still keep the invariants of the class in case of error. In summary, the 
    //exception is handled as expected. Removing catch-all here will break the assumptions (
    //post-conditions) and potentially cause regressions. Instead, a new message is added in case
    //an exception is thrown to be more explicit. A log event is added as part of handling for the 
    //exceptions.
    catch( ... )
    {

        return FALSE;
    }

    m_rstatus_stack.Append( STMRR_SUCCESS );
    return TRUE;
}

bool SimpleTMeshUserData::ReadVersion01( ON_BinaryArchive &file )
{
    int num_blocks = 0;
    if( !file.ReadInt(&num_blocks) )
    {
        m_rstatus_stack.Append( STMRR_ERROR_CORRUPT_BLOCK_STRUCTURE );
        return FALSE;
    }

    for( int i = 0; i < num_blocks; ++i )
    {
        // Identifier for the type of the block.
        ON_String tag;
        if( !file.ReadString(tag) )
        {
            m_rstatus_stack.Append( STMRR_ERROR_CORRUPT_BLOCK_STRUCTURE );
            return FALSE;
        }

        // Types of reads to perform (implies number of reads
        // for this sub-block).
        ON_String types_str;
        if( !file.ReadString(types_str) )
        {
            m_rstatus_stack.Append( STMRR_ERROR_CORRUPT_BLOCK_STRUCTURE );
            return FALSE;
        }

        if( tag == "null" || types_str == "" )
        {
            m_rstatus_stack.Append( STMRR_WARNING_INVALID_BLOCK_TYPES_SPECIFIED );
            continue;
        }
        else if( tag == "tsp01" )
        {
            // This is an old format that's no longer supported.
        }
        else if( tag == "tsm01" )
        {
            if( types_str != "buffer bool" )
                m_rstatus_stack.Append( STMRR_WARNING_INCORRECT_TYPES_FOR_BLOCK );

            // read tsm data
            if( !ReadTsmBlock(file) )
            {
                m_rstatus_stack.Append( STMRR_ERROR_READING_TSM_BLOCK );
                return FALSE;
            }
            continue;
        }
        else if( tag == "cache06" )
        {
            // A rhino-specific cache that's not useful to load outside of rhino
        }
        else if( tag == "cache04" )
        {
            // An old cache format that is unused
        }
        else if( tag == "h1d" )
        {
            // This is the save-limited trial save count, no longer used.
        }
        else if( tag == "trialrecovery" )
        {
            continue;
        }
        else if( tag == "smoothsetting" )
        {
            continue;
        }
        else
        {
            m_rstatus_stack.Append( STMRR_WARNING_UNKNOWN_TAG_TYPE );
        }
    }

    // If I don't have a valid string at the end of this, then it's still not a successful read.
    if( m_tsm_string.Length() == 0 )
    {
        m_rstatus_stack.Append( STMRR_ERROR_NO_TSM_STRING_FOUND );
        return FALSE;
    }

    m_rstatus_stack.Append( STMRR_SUCCESS );
    return TRUE;
}

bool SimpleTMeshUserData::ReadTsmBlock( ON_BinaryArchive &file )
{
    size_t buf_size;
    if( !file.ReadCompressedBufferSize(&buf_size) )
    {
        m_rstatus_stack.Append( STMRR_ERROR_READING_BLOCK_ELEMENTS );
        return FALSE;
    }

    char *buf = new char[buf_size];
    bool bFailedCRC = false;
    if ( !file.ReadCompressedBuffer(buf_size, buf, &bFailedCRC) )
    {
        delete [] buf;
        m_rstatus_stack.Append( STMRR_ERROR_READING_BLOCK_ELEMENTS );
        return FALSE;
    }

    ON_String temp = ON_String( buf, int(buf_size) );
    delete [] buf;

    if( temp.Length() == 0 )
    {
        m_rstatus_stack.Append( STMRR_ERROR_NO_TSM_STRING_FOUND );
        return FALSE;
    }
        
    m_tsm_string = temp;

    if ( !file.ReadBool( &m_saved_in_smooth_mode ) )
    {
        m_rstatus_stack.Append( STMRR_ERROR_READING_BLOCK_ELEMENTS );
        return FALSE;
    }

    return TRUE;
}

bool SimpleTMeshUserData::ReadLegacyBlock( ON_BinaryArchive &file, bool &end_reached )
{
    end_reached = false;

    if ( !file.ReadString( m_tsm_string ) )
        return FALSE;

    // Note: this is a hack.  According to the McNeel wiki:
    //  'If any ON_BinaryArchive::Read*() functions return false than you should 
    //  'immediately return false otherwise return true when all data was read.
    //
    // We are not returning false immediately...this could have unforseen consequences.
    m_saved_in_smooth_mode = false;
    if ( !file.ReadBool( &m_saved_in_smooth_mode ) )
    {
        end_reached = true;
        m_saved_in_smooth_mode = false;
    }

    return TRUE;
}

bool SimpleTMeshUserData::Write( ON_BinaryArchive& file ) const
{
    bool success = TRUE;

    return success;
}
