
#ifndef _RHINO_TSPLINE_USER_DATA_H
#define _RHINO_TSPLINE_USER_DATA_H
class SimpleTMeshUserData : public ON_UserData
{
    ON_OBJECT_DECLARE(SimpleTMeshUserData);
public:
    static GUID m_uuid;

    SimpleTMeshUserData();

    bool GetDescription(ON_wString&) override;

    /// Convenience function to get SimpleTMeshUserData from any ON_Geometry.
    /// Returns nullptr if there was no such user data attached to the geometry
    static const SimpleTMeshUserData* Get(const ON_Geometry* mesh_or_brep);

    bool Archive() const override;

    /// The binary archive scheme has gone through a couple of revisions.
    ///
    /// Up through version 1.4: (See TSUserData.h)
    ///
    /// Type    |  Description
    /// --------------------------------------------------------------
    /// String  |  TSP geometry description.
    ///
    ///
    /// During the version 2.0 Beta:
    ///
    /// Type    |  Description
    /// --------------------------------------------------------------
    /// String  |  TSM geometry description.
    /// Bool    |  Whether the TSM should be in smooth mode
    ///
    ///
    /// Versions 2.0+ and 1.5:
    ///
    /// Type    |  Description
    /// --------------------------------------------------------------
    /// String  |  TSM geometry description (Not used).
    /// Bool    |  Whether the TSM should be in smooth mode (Not used).
    /// Bool    |  Whether additional data is included below
    /// Int     |  Data version number (beginning at 1 for the 2.0 release)
    /// Int     |  Number of data sub-blocks (one for each type of data to store)
    /// (for each sub-block)
    ///  String |  Identification tag for this sub-block ("TSP", for example)
    ///  String |  Type of data to read in this sub-block.  This is a space-separated
    ///            list of the following choices: buffer (compressed ASCII string),
    ///            bool, int, float, double, string.
    ///            The number and types of items in this list determine the following
    ///            sequence of data to read.
    ///  Type0  |  Zero-th data item.
    ///  Type1  |  First data item.
    ///   ...   |    ...
    ///  TypeN  |  Nth data item.
    ///
    ///
    ///
    /// Data Types for version-1 sub-blocks:
    ///
    /// Type    |  ON_BinaryArchive functions / Description
    /// --------------------------------------------------------------
    /// bool    |  ReadBool(bool*), WriteBool(bool)
    ///         |    A Boolean value.
    /// int     |  ReadInt(int*), ReadInt(unsigned int*), WriteInt(int), WriteInt(unsigned int), etc.
    ///         |    An integer value.
    /// float   |  ReadFloat(float*), WriteFloat(float)
    ///         |    A floating-point value.
    /// double  |  ReadDouble(double*), WriteDouble(double)
    ///         |    A double-precision floating-point value.
    /// string  |  ReadString(ON_String&), WriteString(ON_String&), etc.
    ///         |    A string of characters.
    /// buffer  |  ReadCompressedBuffer(size_t [size], void* [buf], int* [failure]), WriteCompressedBuffer(int, void*)
    ///         |    An array of [size] bytes.  [size] is obtained using ReadCompressedBufferSize(size_t*).
    ///         |    [buf] is automatically compressed using zlib 1.1.3, according to the Rhino documentation.
    /// array   |  ReadByte(size_t [size], void* [buf]), WriteByte(size_t [size], const void* [buf])
    ///         |    An array of bytes. Must write the array size before hand as a uint.
    ///         |    (i.e. uint size; file.WriteInt(size); file.WriteByte(size, array);)
    ///         |  Note: This data type has been deprecated, as it was included after the realease of version 2.0.
    ///
    /// Important: Any additional data types (including using the "array" data type) will require the
    /// "Data version number" to be incremented.  This is likely to break backward-compatibility.
    ///
    /// Note that ON_BinaryArchive will force whatever data it reads into the specified
    /// data type.  A potential side-effect is that, for example, data which has been
    /// written via WriteInt(unsigned int) can be successfully read via ReadInt(int*),
    /// or, in fact, via any other Read* function, resulting in data corruption.

    bool Write(
        ON_BinaryArchive&  // serialize definition to binary archive
    ) const override;

    /// See the comment for SimpleTMeshUserData::Write( ON_BinaryArchive& ) const.
    bool Read(
        ON_BinaryArchive&  // restore definition from binary archive
    ) override;

public:
    enum ReadResult { STMRR_SUCCESS = 0, STMRR_ERROR_COULDNT_READ_LEGACY_BLOCK, STMRR_ERROR_UNKNOWN_BLOCK_STRUCTURE, STMRR_ERROR_READING_BLOCK_STRUCTURE, STMRR_ERROR_CORRUPT_BLOCK_STRUCTURE, STMRR_ERROR_CAUGHT_EXCEPTION, STMRR_WARNING_UNABLE_TO_READ_SMOOTH_SETTING, STMRR_WARNING_INVALID_BLOCK_TYPES_SPECIFIED, STMRR_WARNING_INCORRECT_TYPES_FOR_BLOCK, STMRR_INFO_RECOVERED_FROM_TRIAL, STMRR_ERROR_READING_TSM_BLOCK, STMRR_ERROR_READING_TRIAL_BLOCK, STMRR_WARNING_UNABLE_TO_READ_SMOOTH_SETTING_BLOCK, STMRR_WARNING_UNKNOWN_TAG_TYPE, STMRR_ERROR_READING_BLOCK_ELEMENTS, STMRR_ERROR_UNKNOWN_DATA_TYPE, STMRR_ERROR_NO_TSM_STRING_FOUND };
    enum WriteResult { STMWR_SUCCESS = 0, STMWR_ERROR_NO_TSM_STRING, STMWR_ERROR_CAUGHT_EXCEPTION, STMWR_ERROR_WRITING_TSM_BLOCK, STMWR_ERROR_WRITING_SMOOTH_SETTING_BLOCK, STMWR_FAILURE };

    mutable ON_SimpleArray< ReadResult > m_rstatus_stack;
    mutable ON_SimpleArray< WriteResult > m_wstatus_stack;

    // Where the TSM format string is stored.
    ON_String m_tsm_string;
    bool m_saved_in_smooth_mode;

protected:
    bool ReadVersion01(ON_BinaryArchive& file);
    bool ReadTsmBlock(ON_BinaryArchive& file);
    //bool ReadTrialRecoveryBlock(ON_BinaryArchive& file);
    bool ReadLegacyBlock(ON_BinaryArchive& file, bool& end_reached);
    //bool ReadSmoothSettingBlock(ON_BinaryArchive& file);

    //bool WriteTsmBlock(ON_BinaryArchive& file) const;
    //bool WriteCacheBlock(ON_BinaryArchive& file) const;
    //bool WriteSmoothSettingBlock(ON_BinaryArchive& file) const;
};
#endif // _RHINO_TSPLINE_USER_DATA_H